/* SPDX-License-Identifier: GPL-2.0 */
#ifndef __ASM_BARRIER_H
#define __ASM_BARRIER_H

#ifndef __ASSEMBLY__

#define nop() __asm__ __volatile__("mov\tr0,r0\t@ nop\n\t");

#if __LINUX_ARM_ARCH__ >= 7 ||		\
	(__LINUX_ARM_ARCH__ == 6 && defined(CONFIG_CPU_32v6K))
#define sev()	__asm__ __volatile__ ("sev" : : : "memory")
#define wfe()	__asm__ __volatile__ ("wfe" : : : "memory")
#define wfi()	__asm__ __volatile__ ("wfi" : : : "memory")
#else
#define wfe()	do { } while (0)
#endif

#if __LINUX_ARM_ARCH__ >= 7
#define isb(option) __asm__ __volatile__ ("isb " #option : : : "memory")
#define dsb(option) __asm__ __volatile__ ("dsb " #option : : : "memory")
#define dmb(option) __asm__ __volatile__ ("dmb " #option : : : "memory")
#ifdef CONFIG_THUMB2_KERNEL
#define CSDB	".inst.w 0xf3af8014"
#else
#define CSDB	".inst	0xe320f014"
#endif
#define csdb() __asm__ __volatile__(CSDB : : : "memory")
#elif defined(CONFIG_CPU_XSC3) || __LINUX_ARM_ARCH__ == 6
#define isb(x) __asm__ __volatile__ ("mcr p15, 0, %0, c7, c5, 4" \
				    : : "r" (0) : "memory")
#define dsb(x) __asm__ __volatile__ ("mcr p15, 0, %0, c7, c10, 4" \
				    : : "r" (0) : "memory")
#define dmb(x) __asm__ __volatile__ ("mcr p15, 0, %0, c7, c10, 5" \
				    : : "r" (0) : "memory")
#elif defined(CONFIG_CPU_FA526)
#define isb(x) __asm__ __volatile__ ("mcr p15, 0, %0, c7, c5, 4" \
				    : : "r" (0) : "memory")
#define dsb(x) __asm__ __volatile__ ("mcr p15, 0, %0, c7, c10, 4" \
				    : : "r" (0) : "memory")
#define dmb(x) __asm__ __volatile__ ("" : : : "memory")
#else
#define isb(x) __asm__ __volatile__ ("" : : : "memory")
#define dsb(x) __asm__ __volatile__ ("mcr p15, 0, %0, c7, c10, 4" \
				    : : "r" (0) : "memory")
#define dmb(x) __asm__ __volatile__ ("" : : : "memory")
#endif

#ifndef CSDB
#define CSDB
#endif
#ifndef csdb
#define csdb()
#endif

#ifdef CONFIG_ARM_HEAVY_MB
extern void (*soc_mb)(void);
extern void arm_heavy_mb(void);
#define __arm_heavy_mb(x...) do { dsb(x); arm_heavy_mb(); } while (0)
#else
#define __arm_heavy_mb(x...) dsb(x)
#endif

#if defined(CONFIG_ARM_DMA_MEM_BUFFERABLE) || defined(CONFIG_SMP)
#define mb()		__arm_heavy_mb()
#define rmb()		dsb()
#define wmb()		__arm_heavy_mb(st)
#define dma_rmb()	dmb(osh)
#define dma_wmb()	dmb(oshst)
#else
#define mb()		barrier()
#define rmb()		barrier()
#define wmb()		barrier()
#define dma_rmb()	barrier()
#define dma_wmb()	barrier()
#endif

#define __smp_mb()	dmb(ish)
#define __smp_rmb()	__smp_mb()
#define __smp_wmb()	dmb(ishst)

#ifdef CONFIG_CPU_SPECTRE
/*
 * Generate a mask for array_index_nospec() that is ~0UL when 0 <= index < size
 * and 0 otherwise.
 */
static __respectre_inline unsigned long array_index_mask_nospec(unsigned long index,
						    unsigned long size)
{
	unsigned long mask;

	asm volatile(
		"cmp	%[index], %[size]\n"
	"	sbc	%[mask], %[index], %[index]\n"
	CSDB
	: [mask] "=r" (mask)
	: [index] "r" (index), [size] "Ir" (size)
	: "cc");

	return mask;
}
#define array_index_mask_nospec array_index_mask_nospec

static __respectre_inline unsigned long array_index_mask_nospec_extra(unsigned long index,
						    unsigned long size)
{
	unsigned long mask;

	if (__builtin_constant_p(size))
		return size ? array_index_mask_nospec(index, size) : ~0UL;

	asm volatile(
		"cmp	%[index], %[size]\n"
	"	sbc	%[mask], %[index], %[index]\n"
	"	cmp	%[size], #1\n"
	"	sbc	%[mask], %[mask], #0\n"
	CSDB
	: [mask] "=&r" (mask)
	: [index] "r" (index), [size] "r" (size)
	: "cc");

	return mask;
}
#define array_index_mask_nospec_extra array_index_mask_nospec_extra

static __respectre_inline unsigned long long array_index_mask_nospec_ll(unsigned long long index,
						    unsigned long long size)
{
	unsigned long mask;
	unsigned long hindex = index >> 32, lindex = index;
	unsigned long hsize = size >> 32, lsize = size;

	asm volatile(
		"cmp	%[lindex], %[lsize]\n"
	"	sbc	%[mask], %[hindex], %[hsize]\n"
	"	sbc	%[mask], %[lindex], %[lindex]\n"
	CSDB
	: [mask] "=&r" (mask)
	: [lindex] "r" (lindex), [hindex] "r" (hindex),
	  [lsize] "Ir" (lsize), [hsize] "Ir" (hsize)
	: "cc");

	return mask | (unsigned long long)mask << 32;
}
#define array_index_mask_nospec_ll array_index_mask_nospec_ll

static __respectre_inline unsigned long long array_index_mask_nospec_extra_ll(unsigned long long index,
						    unsigned long long size)
{
	unsigned long mask;
	unsigned long hindex = index >> 32, lindex = index;
	unsigned long hsize = size >> 32, lsize = size;

	if (__builtin_constant_p(size))
		return size ? array_index_mask_nospec_ll(index, size) : ~0ULL;
	if (__builtin_constant_p(hsize))
		return array_index_mask_nospec_ll(index, size);

	asm volatile(
		"cmp	%[lindex], %[lsize]\n"
	"	sbc	%[mask], %[hindex], %[hsize]\n"
	"	sbc	%[mask], %[lindex], %[lindex]\n"
	"	cmp	%[hsize], #1\n"
	"	sbc	%[mask], %[mask], #0\n"
	CSDB
	: [mask] "=&r" (mask)
	: [lindex] "r" (lindex), [hindex] "r" (hindex),
	  [lsize] "Ir" (lsize), [hsize] "r" (hsize)
	: "cc");

	return mask | (unsigned long long)mask << 32;
}
#define array_index_mask_nospec_extra_ll array_index_mask_nospec_extra_ll
#endif

static __respectre_inline void barrier_nospec(void)
{
	csdb();
}
#define barrier_nospec barrier_nospec

#include <asm-generic/barrier.h>

#endif /* !__ASSEMBLY__ */
#endif /* __ASM_BARRIER_H */
