/* SPDX-License-Identifier: GPL-2.0 */
/* atomic.h: Thankfully the V9 is at least reasonable for this
 *           stuff.
 *
 * Copyright (C) 1996, 1997, 2000, 2012 David S. Miller (davem@redhat.com)
 */

#ifndef __ARCH_SPARC64_ATOMIC__
#define __ARCH_SPARC64_ATOMIC__

#include <linux/types.h>
#include <asm/cmpxchg.h>
#include <asm/barrier.h>

#define ATOMIC_INIT(i)		{ (i) }
#define ATOMIC64_INIT(i)	{ (i) }

#define atomic_read(v)		READ_ONCE((v)->counter)
#define atomic_read_unchecked	atomic_read
#define atomic64_read(v)	READ_ONCE((v)->counter)
#define atomic64_read_unchecked	atomic64_read

#define atomic_set(v, i)	WRITE_ONCE(((v)->counter), (i))
#define atomic_set_unchecked	atomic_set
#define atomic64_set(v, i)	WRITE_ONCE(((v)->counter), (i))
#define atomic64_set_unchecked	atomic64_set

#define __ATOMIC_OP(op, suffix)						\
void atomic_##op##suffix(int, atomic##suffix##_t *);			\
void atomic64_##op##suffix(s64, atomic64##suffix##_t *);

#define ATOMIC_OP(op) __ATOMIC_OP(op, ) __ATOMIC_OP(op, _unchecked)

#define __ATOMIC_OP_RETURN(op, suffix)					\
int atomic_##op##_return##suffix(int, atomic##suffix##_t *);		\
s64 atomic64_##op##_return##suffix(s64, atomic64##suffix##_t *);

#define ATOMIC_OP_RETURN(op) __ATOMIC_OP_RETURN(op, ) __ATOMIC_OP_RETURN(op, _unchecked)

#define ATOMIC_FETCH_OP(op)						\
int atomic_fetch_##op(int, atomic_t *);					\
s64 atomic64_fetch_##op(s64, atomic64_t *);

#define ATOMIC_OPS(op) ATOMIC_OP(op) ATOMIC_OP_RETURN(op) ATOMIC_FETCH_OP(op)

ATOMIC_OPS(add)
ATOMIC_OPS(sub)

#undef ATOMIC_OPS
#define ATOMIC_OPS(op) ATOMIC_OP(op) ATOMIC_FETCH_OP(op)

ATOMIC_OPS(and)
ATOMIC_OPS(or)
ATOMIC_OPS(xor)

#undef ATOMIC_OPS
#undef ATOMIC_FETCH_OP
#undef ATOMIC_OP_RETURN
#undef __ATOMIC_OP_RETURN
#undef ATOMIC_OP
#undef __ATOMIC_OP

#define atomic_cmpxchg(v, o, n) (cmpxchg(&((v)->counter), (o), (n)))
static inline int atomic_cmpxchg_unchecked(atomic_unchecked_t *v, int old, int new)
{
	return cmpxchg(&v->counter, old, new);
}

static inline int atomic_xchg(atomic_t *v, int new)
{
	return xchg(&v->counter, new);
}

static inline int atomic_xchg_unchecked(atomic_unchecked_t *v, int new)
{
	return xchg(&v->counter, new);
}

#define atomic_fetch_add_unless atomic_fetch_add_unless
static inline s64 atomic_fetch_add_unless(atomic_t *v, int a, int u)
{
	int c, old, new;
	c = atomic_read(v);
	for (;;) {
		if (unlikely(c == u))
			break;

		asm volatile("addcc %2, %0, %0\n"

#ifdef CONFIG_PAX_REFCOUNT
			     "tvs %%icc, 6\n"
#endif

			     : "=r" (new)
			     : "0" (c), "ir" (a)
			     : "cc");

		old = atomic_cmpxchg(v, c, new);
		if (likely(old == c))
			break;
		c = old;
	}
	return c;
}

#define atomic64_cmpxchg(v, o, n) \
	((__typeof__((v)->counter))cmpxchg(&((v)->counter), (o), (n)))
static inline s64 atomic64_cmpxchg_unchecked(atomic64_unchecked_t *v, s64 old,
					      s64 new)
{
	return cmpxchg(&(v->counter), old, new);
}

#define atomic64_xchg(v, new) (xchg(&((v)->counter), new))
static inline s64 atomic64_xchg_unchecked(atomic64_unchecked_t *v, s64 new)
{
	return xchg(&v->counter, new);
}

#define atomic64_fetch_add_unless atomic64_fetch_add_unless
static inline s64 atomic64_fetch_add_unless(atomic64_t *v, s64 a, s64 u)
{
	s64 c, old, new;
	c = atomic64_read(v);
	for (;;) {
		if (unlikely(c == u))
			break;

		asm volatile("addcc %2, %0, %0\n"

#ifdef CONFIG_PAX_REFCOUNT
			     "tvs %%xcc, 6\n"
#endif

			     : "=r" (new)
			     : "0" (c), "ir" (a)
			     : "cc");

		old = atomic64_cmpxchg(v, c, new);
		if (likely(old == c))
			break;
		c = old;
	}
	return c;
}

s64 atomic64_dec_if_positive(atomic64_t *v);
#define atomic64_dec_if_positive atomic64_dec_if_positive

#endif /* !(__ARCH_SPARC64_ATOMIC__) */
